from __future__ import annotations
from collections import defaultdict

from datetime import datetime, timedelta, timezone
import enum
import os
from pathlib import Path
from textwrap import indent
from typing import Dict, List, Optional, Tuple

from hpcflow.sdk import app
from hpcflow.sdk.core.element import ElementResources
from hpcflow.sdk.core.errors import (
    JobscriptSubmissionFailure,
    MissingEnvironmentError,
    MissingEnvironmentExecutableError,
    MissingEnvironmentExecutableInstanceError,
    SubmissionFailure,
)
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike


def timedelta_format(td: timedelta) -> str:
    days, seconds = td.days, td.seconds
    hours = seconds // (60 * 60)
    seconds -= hours * (60 * 60)
    minutes = seconds // 60
    seconds -= minutes * 60
    return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"


def timedelta_parse(td_str: str) -> timedelta:
    days, other = td_str.split("-")
    days = int(days)
    hours, mins, secs = [int(i) for i in other.split(":")]
    return timedelta(days=days, hours=hours, minutes=mins, seconds=secs)


class SubmissionStatus(enum.Enum):
    PENDING = 0  # not yet submitted
    SUBMITTED = 1  # all jobscripts submitted successfully
    PARTIALLY_SUBMITTED = 2  # some jobscripts submitted successfully


class Submission(JSONLike):
    _child_objects = (
        ChildObjectSpec(
            name="jobscripts",
            class_name="Jobscript",
            is_multiple=True,
            parent_ref="_submission",
        ),
        ChildObjectSpec(
            name="environments",
            class_name="EnvironmentsList",
        ),
    )

    def __init__(
        self,
        index: int,
        jobscripts: List[app.Jobscript],
        workflow: Optional[app.Workflow] = None,
        submission_parts: Optional[Dict] = None,
        JS_parallelism: Optional[bool] = None,
        environments: Optional[app.EnvironmentsList] = None,
    ):
        self._index = index
        self._jobscripts = jobscripts
        self._submission_parts = submission_parts or {}
        self._JS_parallelism = JS_parallelism
        self._environments = environments

        self._submission_parts_lst = None  # assigned on first access; datetime objects

        if workflow:
            self.workflow = workflow

        self._set_parent_refs()

        for js_idx, js in enumerate(self.jobscripts):
            js._index = js_idx

    def _set_environments(self):
        filterable = ElementResources.get_env_instance_filterable_attributes()

        # map required environments and executable labels to job script indices:
        req_envs = defaultdict(lambda: defaultdict(set))
        for js_idx, js_i in enumerate(self.jobscripts):
            for run in js_i.all_EARs:
                env_label = run.action.get_environment_label()
                for exec_label_j in run.action.get_required_executables():
                    req_envs[env_label][exec_label_j].add(js_idx)
                if env_label not in req_envs:
                    req_envs[env_label] = {}

        # check these envs/execs exist in app data:
        envs = []
        for env_lab, exec_js in req_envs.items():
            try:
                env_i = self.app.envs.get(env_lab)
            except ValueError:
                raise MissingEnvironmentError(
                    f"The environment {env_lab!r} is not defined on this machine, so the "
                    f"submission cannot be created."
                ) from None
            else:
                envs.append(env_i)

            for exec_i_lab, js_idx_set in exec_js.items():
                try:
                    exec_i = env_i.executables.get(exec_i_lab)
                except ValueError:
                    raise MissingEnvironmentExecutableError(
                        f"The environment {env_lab!r} as defined on this machine has no "
                        f"executable labelled {exec_i_lab!r}, which is required for this "
                        f"submission, so the submission cannot be created."
                    ) from None

                # check matching executable instances exist:
                for js_idx_j in js_idx_set:
                    js_j = self.jobscripts[js_idx_j]
                    filter_exec = {j: getattr(js_j.resources, j) for j in filterable}
                    exec_instances = exec_i.filter_instances(**filter_exec)
                    if not exec_instances:
                        raise MissingEnvironmentExecutableInstanceError(
                            f"No matching executable instances found for executable "
                            f"{exec_i_lab!r} of environment {env_lab!r} for jobscript "
                            f"index {js_idx_j!r} with requested resources "
                            f"{filter_exec!r}."
                        )

        # save env definitions to the environments attribute:
        self._environments = self.app.EnvironmentsList(envs)

    def to_dict(self):
        dct = super().to_dict()
        del dct["_workflow"]
        del dct["_index"]
        del dct["_submission_parts_lst"]
        dct = {k.lstrip("_"): v for k, v in dct.items()}
        return dct

    @property
    def index(self) -> int:
        return self._index

    @property
    def environments(self) -> app.EnvironmentsList:
        return self._environments

    @property
    def submission_parts(self) -> List[Dict]:
        if not self._submission_parts:
            return []

        if self._submission_parts_lst is None:
            self._submission_parts_lst = [
                {
                    "submit_time": datetime.strptime(dt, self.workflow.ts_fmt)
                    .replace(tzinfo=timezone.utc)
                    .astimezone(),
                    "jobscripts": js_idx,
                }
                for dt, js_idx in self._submission_parts.items()
            ]
        return self._submission_parts_lst

    def get_start_time(self, submit_time: str) -> Union[datetime, None]:
        """Get the start time of a given submission part."""
        js_idx = self._submission_parts[submit_time]
        all_part_starts = []
        for i in js_idx:
            if self.jobscripts[i].start_time:
                all_part_starts.append(self.jobscripts[i].start_time)
        if all_part_starts:
            return min(all_part_starts)
        else:
            return None

    def get_end_time(self, submit_time: str) -> Union[datetime, None]:
        """Get the end time of a given submission part."""
        js_idx = self._submission_parts[submit_time]
        all_part_ends = []
        for i in js_idx:
            if self.jobscripts[i].end_time:
                all_part_ends.append(self.jobscripts[i].end_time)
        if all_part_ends:
            return max(all_part_ends)
        else:
            return None

    @property
    def start_time(self):
        """Get the first non-None start time over all submission parts."""
        all_start_times = []
        for submit_time in self._submission_parts:
            start_i = self.get_start_time(submit_time)
            if start_i:
                all_start_times.append(start_i)
        if all_start_times:
            return max(all_start_times)
        else:
            return None

    @property
    def end_time(self):
        """Get the final non-None end time over all submission parts."""
        all_end_times = []
        for submit_time in self._submission_parts:
            end_i = self.get_end_time(submit_time)
            if end_i:
                all_end_times.append(end_i)
        if all_end_times:
            return max(all_end_times)
        else:
            return None

    @property
    def jobscripts(self) -> List:
        return self._jobscripts

    @property
    def JS_parallelism(self):
        return self._JS_parallelism

    @property
    def workflow(self) -> List:
        return self._workflow

    @workflow.setter
    def workflow(self, wk):
        self._workflow = wk

    @property
    def jobscript_indices(self) -> Tuple[int]:
        """All associated jobscript indices."""
        return tuple(i.index for i in self.jobscripts)

    @property
    def submitted_jobscripts(self) -> Tuple[int]:
        """Jobscript indices that have been successfully submitted."""
        return tuple(j for i in self.submission_parts for j in i["jobscripts"])

    @property
    def outstanding_jobscripts(self) -> Tuple[int]:
        """Jobscript indices that have not yet been successfully submitted."""
        return tuple(set(self.jobscript_indices) - set(self.submitted_jobscripts))

    @property
    def status(self):
        if not self.submission_parts:
            return SubmissionStatus.PENDING
        else:
            if set(self.submitted_jobscripts) == set(self.jobscript_indices):
                return SubmissionStatus.SUBMITTED
            else:
                return SubmissionStatus.PARTIALLY_SUBMITTED

    @property
    def needs_submit(self):
        return self.status in (
            SubmissionStatus.PENDING,
            SubmissionStatus.PARTIALLY_SUBMITTED,
        )

    @property
    def path(self):
        return self.workflow.submissions_path / str(self.index)

    @property
    def all_EAR_IDs(self):
        return [i for js in self.jobscripts for i in js.all_EAR_IDs]

    @property
    def all_EARs(self):
        return [i for js in self.jobscripts for i in js.all_EARs]

    @property
    def EARs_by_elements(self):
        task_elem_EARs = defaultdict(lambda: defaultdict(list))
        for i in self.all_EARs:
            task_elem_EARs[i.task.index][i.element.index].append(i)
        return task_elem_EARs

    @property
    def abort_EARs_file_name(self):
        return f"abort_EARs.txt"

    @property
    def abort_EARs_file_path(self):
        return self.path / self.abort_EARs_file_name

    def get_active_jobscripts(
        self, as_json: bool = False
    ) -> List[Tuple[int, Dict[int, JobscriptElementState]]]:
        """Get jobscripts that are active on this machine, and their active states."""
        # this returns: {JS_IDX: {JS_ELEMENT_IDX: STATE}}
        # TODO: query the scheduler once for all jobscripts?
        out = {}
        for js in self.jobscripts:
            active_states = js.get_active_states(as_json=as_json)
            if active_states:
                out[js.index] = active_states
        return out

    def _write_abort_EARs_file(self):
        with self.abort_EARs_file_path.open(mode="wt", newline="\n") as fp:
            # write a single line for each EAR currently in the workflow:
            fp.write("\n".join("0" for _ in range(self.workflow.num_EARs)) + "\n")

    def _set_run_abort(self, run_ID: int):
        """Modify the abort runs file to indicate a specified run should be aborted."""
        with self.abort_EARs_file_path.open(mode="rt", newline="\n") as fp:
            lines = fp.read().splitlines()
        lines[run_ID] = "1"

        # write a new temporary run-abort file:
        tmp_suffix = self.abort_EARs_file_path.suffix + ".tmp"
        tmp = self.abort_EARs_file_path.with_suffix(tmp_suffix)
        self.app.submission_logger.debug(f"Creating temporary run abort file: {tmp!r}.")
        with tmp.open(mode="wt", newline="\n") as fp:
            fp.write("\n".join(i for i in lines) + "\n")

        # atomic rename, overwriting original:
        self.app.submission_logger.debug(
            "Replacing original run abort file with new temporary file."
        )
        os.replace(src=tmp, dst=self.abort_EARs_file_path)

    @staticmethod
    def get_unique_schedulers_of_jobscripts(
        jobscripts: List[Jobscript],
    ) -> Dict[Tuple[Tuple[int, int]], Scheduler]:
        """Get unique schedulers and which of the passed jobscripts they correspond to.

        Uniqueness is determines only by the `Scheduler.unique_properties` tuple.

        """
        js_idx = []
        schedulers = []

        # list of tuples of scheduler properties we consider to determine "uniqueness",
        # with the first string being the scheduler type (class name):
        seen_schedulers = []

        for js in jobscripts:
            if js.scheduler.unique_properties not in seen_schedulers:
                seen_schedulers.append(js.scheduler.unique_properties)
                schedulers.append(js.scheduler)
                js_idx.append([])
            sched_idx = seen_schedulers.index(js.scheduler.unique_properties)
            js_idx[sched_idx].append((js.submission.index, js.index))

        sched_js_idx = dict(zip((tuple(i) for i in js_idx), schedulers))

        return sched_js_idx

    def get_unique_schedulers(self) -> Dict[Tuple[int], Scheduler]:
        """Get unique schedulers and which of this submission's jobscripts they
        correspond to."""
        return self.get_unique_schedulers_of_jobscripts(self.jobscripts)

    def get_unique_shells(self) -> Dict[Tuple[int], Shell]:
        """Get unique shells and which jobscripts they correspond to."""
        js_idx = []
        shells = []

        for js in self.jobscripts:
            if js.shell not in shells:
                shells.append(js.shell)
                js_idx.append([])
            shell_idx = shells.index(js.shell)
            js_idx[shell_idx].append(js.index)

        shell_js_idx = dict(zip((tuple(i) for i in js_idx), shells))

        return shell_js_idx

    def _raise_failure(self, submitted_js_idx, exceptions):
        msg = f"Some jobscripts in submission index {self.index} could not be submitted"
        if submitted_js_idx:
            msg += f" (but jobscripts {submitted_js_idx} were submitted successfully):"
        else:
            msg += ":"

        msg += "\n"
        for sub_err in exceptions:
            msg += (
                f"Jobscript {sub_err.js_idx} at path: {str(sub_err.js_path)!r}\n"
                f"Submit command: {sub_err.submit_cmd!r}.\n"
                f"Reason: {sub_err.message!r}\n"
            )
            if sub_err.subprocess_exc is not None:
                msg += f"Subprocess exception: {sub_err.subprocess_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Subprocess job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.stdout:
                msg += f"Submission stdout:\n{indent(sub_err.stdout, '  ')}\n"
            if sub_err.stderr:
                msg += f"Submission stderr:\n{indent(sub_err.stderr, '  ')}\n"

        raise SubmissionFailure(message=msg)

    def _append_submission_part(self, submit_time: str, submitted_js_idx: List[int]):
        self._submission_parts[submit_time] = submitted_js_idx
        self.workflow._store.add_submission_part(
            sub_idx=self.index,
            dt_str=submit_time,
            submitted_js_idx=submitted_js_idx,
        )

    def submit(
        self,
        status,
        ignore_errors: Optional[bool] = False,
        print_stdout: Optional[bool] = False,
        add_to_known: Optional[bool] = True,
    ) -> List[int]:
        """Generate and submit the jobscripts of this submission."""

        # if JS_parallelism explicitly requested but store doesn't support, raise:
        supports_JS_para = self.workflow._store._features.jobscript_parallelism
        if self.JS_parallelism:
            if not supports_JS_para:
                status.stop()
                raise ValueError(
                    f"Store type {self.workflow._store!r} does not support jobscript "
                    f"parallelism."
                )
        elif self.JS_parallelism is None:
            self._JS_parallelism = supports_JS_para

        # set os_name and shell_name for each jobscript:
        for js in self.jobscripts:
            js._set_os_name()
            js._set_shell_name()
            js._set_scheduler_name()

        outstanding = self.outstanding_jobscripts

        # get scheduler, shell and OS version information (also an opportunity to fail
        # before trying to submit jobscripts):
        js_vers_info = {}
        for js_indices, sched in self.get_unique_schedulers().items():
            try:
                vers_info = sched.get_version_info()
            except Exception as err:
                if ignore_errors:
                    vers_info = {}
                else:
                    raise err
            for _, js_idx in js_indices:
                if js_idx in outstanding:
                    if js_idx not in js_vers_info:
                        js_vers_info[js_idx] = {}
                    js_vers_info[js_idx].update(vers_info)

        for js_indices, shell in self.get_unique_shells().items():
            try:
                vers_info = shell.get_version_info()
            except Exception as err:
                if ignore_errors:
                    vers_info = {}
                else:
                    raise err
            for js_idx in js_indices:
                if js_idx in outstanding:
                    if js_idx not in js_vers_info:
                        js_vers_info[js_idx] = {}
                    js_vers_info[js_idx].update(vers_info)

        for js_idx, vers_info_i in js_vers_info.items():
            self.jobscripts[js_idx]._set_version_info(vers_info_i)

        # for direct submission, it's important that os_name/shell_name/scheduler_name
        # are made persistent now, because `Workflow.write_commands`, which might be
        # invoked in a new process before submission has completed, needs to know these:
        self.workflow._store._pending.commit_all()

        # TODO: a submission should only be "submitted" once shouldn't it?
        # no; there could be an IO error (e.g. internet connectivity), so might
        # need to be able to reattempt submission of outstanding jobscripts.
        self.path.mkdir(exist_ok=True)
        if not self.abort_EARs_file_path.is_file():
            self._write_abort_EARs_file()

        # map jobscript `index` to (scheduler job ID or process ID, is_array):
        scheduler_refs = {}
        submitted_js_idx = []
        errs = []
        for js in self.jobscripts:
            # check not previously submitted:
            if js.index not in outstanding:
                continue

            # check all dependencies were submitted now or previously:
            if not all(
                i in submitted_js_idx or i in self.submitted_jobscripts
                for i in js.dependencies
            ):
                continue

            try:
                status.update(f"Submitting jobscript {js.index}...")
                js_ref_i = js.submit(scheduler_refs, print_stdout=print_stdout)
                scheduler_refs[js.index] = (js_ref_i, js.is_array)
                submitted_js_idx.append(js.index)

            except JobscriptSubmissionFailure as err:
                errs.append(err)
                continue

        if submitted_js_idx:
            dt_str = datetime.utcnow().strftime(self.app._submission_ts_fmt)
            self._append_submission_part(
                submit_time=dt_str,
                submitted_js_idx=submitted_js_idx,
            )
            # add a record of the submission part to the known-submissions file
            if add_to_known:
                self.app._add_to_known_submissions(
                    wk_path=self.workflow.path,
                    wk_id=self.workflow.id_,
                    sub_idx=self.index,
                    sub_time=dt_str,
                )

        if errs and not ignore_errors:
            status.stop()
            self._raise_failure(submitted_js_idx, errs)

        len_js = len(submitted_js_idx)
        print(f"Submitted {len_js} jobscript{'s' if len_js > 1 else ''}.")

        return submitted_js_idx

    def cancel(self):
        act_js = list(self.get_active_jobscripts())
        if not act_js:
            print("No active jobscripts to cancel.")
            return
        for js_indices, sched in self.get_unique_schedulers().items():
            # filter by active jobscripts:
            js_idx = [i[1] for i in js_indices if i[1] in act_js]
            if js_idx:
                print(
                    f"Cancelling jobscripts {js_idx!r} of submission {self.index} of "
                    f"workflow {self.workflow.name!r}."
                )
                jobscripts = [self.jobscripts[i] for i in js_idx]
                sched_refs = [i.scheduler_js_ref for i in jobscripts]
                sched.cancel_jobs(js_refs=sched_refs, jobscripts=jobscripts)
            else:
                print("No active jobscripts to cancel.")



from hpcflow.app import app as hf


def test_compose_commands_no_shell_var(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        actions=[hf.Action(commands=[hf.Command(command="Start-Sleep 10")])],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts)],
    )
    sub = wk.add_submission()
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(jobscript=js, JS_action_idx=0)
    assert shell_vars == {0: []}


def test_compose_commands_single_shell_var(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 101})],
    )
    sub = wk.add_submission()
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(jobscript=js, JS_action_idx=0)
    assert shell_vars == {0: [("outputs.p1", "parameter_p1", "stdout")]}


def test_compose_commands_multi_single_shell_var(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test_compose_commands",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command="Start-Sleep 10"),
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_compose_commands",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 101})],
    )
    sub = wk.add_submission()
    js = sub.jobscripts[0]
    run = wk.tasks[0].elements[0].iterations[0].action_runs[0]
    _, shell_vars = run.compose_commands(jobscript=js, JS_action_idx=0)
    assert shell_vars == {0: [], 1: [("outputs.p1", "parameter_p1", "stdout")]}



import os
import pytest

from hpcflow.app import app as hf


@pytest.mark.integration
@pytest.mark.parametrize("exit_code", [0, 1, 98, -1, -123124])
def test_action_exit_code_parsing(null_config, tmp_path, exit_code):
    act = hf.Action(commands=[hf.Command(command=f"exit {exit_code}")])
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[act],
    )
    t1 = hf.Task(schema=[s1])
    wk = hf.Workflow.from_template_data(tasks=[t1], template_name="test", path=tmp_path)
    wk.submit(wait=True, add_to_known=False)
    recorded_exit = wk.get_EARs_from_IDs([0])[0].exit_code
    if os.name == "posix":
        # exit code from bash wraps around:
        exit_code %= 256
    assert recorded_exit == exit_code



import numpy as np
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import P1_parameter_cls as P1


@pytest.fixture
def null_config(tmp_path):
    if not hf.is_config_loaded:
        hf.load_config(config_dir=tmp_path)


def test_inputs_value_sequence_label_added_to_path():
    seq = hf.ValueSequence(path="inputs.p1.a", values=[0, 1], nesting_order=0, label=0)
    assert seq.path == "inputs.p1[0].a"


def test_inputs_value_sequence_no_label_added_to_path():
    seq = hf.ValueSequence(path="inputs.p1.a", values=[0, 1], nesting_order=0, label="")
    assert seq.path == "inputs.p1.a"


def test_inputs_value_sequence_label_attr_added():
    seq = hf.ValueSequence(path="inputs.p1[1].a", values=[0, 1], nesting_order=0)
    assert seq.label == "1"


def test_inputs_value_sequence_label_path_unmodified():
    path = "inputs.p1[1].a"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0)
    assert seq.path == path


def test_raise_on_inputs_value_sequence_label_path_unequal():
    with pytest.raises(ValueError):
        hf.ValueSequence(path="inputs.p1[1].a", values=[0, 1], nesting_order=0, label="2")


def test_no_raise_on_inputs_value_sequence_label_path_equal():
    hf.ValueSequence(path="inputs.p1[1].a", values=[0, 1], nesting_order=0, label="1")


def test_no_raise_on_inputs_value_sequence_label_path_cast_equal():
    hf.ValueSequence(path="inputs.p1[1].a", values=[0, 1], nesting_order=0, label=1)


def test_raise_on_resources_value_sequence_with_path_label():
    with pytest.raises(ValueError):
        hf.ValueSequence(path="resources.main[1]", values=[0, 1], nesting_order=0)


def test_raise_on_resources_value_sequence_with_label_arg():
    with pytest.raises(ValueError):
        hf.ValueSequence(path="resources.main", values=[0, 1], nesting_order=0, label=1)


def test_inputs_value_sequence_simple_path_attributes():
    path = "inputs.p1"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0)
    assert seq.path == path
    assert seq.labelled_type == "p1"
    assert seq.normalised_path == "inputs.p1"
    assert seq.normalised_inputs_path == "p1"
    assert seq.path_type == "inputs"
    assert seq.input_type == "p1"
    assert seq.input_path == ""
    assert seq.resource_scope is None


def test_inputs_value_sequence_path_attributes():
    path = "inputs.p1.a.b"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0)
    assert seq.path == path
    assert seq.labelled_type == "p1"
    assert seq.normalised_path == "inputs.p1.a.b"
    assert seq.normalised_inputs_path == "p1.a.b"
    assert seq.path_type == "inputs"
    assert seq.input_type == "p1"
    assert seq.input_path == "a.b"
    assert seq.resource_scope is None


def test_inputs_value_sequence_with_path_label_path_attributes():
    path = "inputs.p1[1].a.b"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0)
    assert seq.path == path
    assert seq.labelled_type == "p1[1]"
    assert seq.normalised_path == "inputs.p1[1].a.b"
    assert seq.normalised_inputs_path == "p1[1].a.b"
    assert seq.path_type == "inputs"
    assert seq.input_type == "p1"
    assert seq.input_path == "a.b"
    assert seq.resource_scope is None


def test_inputs_value_sequence_with_arg_label_path_attributes():
    path = "inputs.p1.a.b"
    new_path = "inputs.p1[1].a.b"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0, label=1)
    assert seq.path == new_path
    assert seq.labelled_type == "p1[1]"
    assert seq.normalised_path == "inputs.p1[1].a.b"
    assert seq.normalised_inputs_path == "p1[1].a.b"
    assert seq.path_type == "inputs"
    assert seq.input_type == "p1"
    assert seq.input_path == "a.b"
    assert seq.resource_scope is None


def test_resources_value_sequence_path_attributes():
    path = "resources.main.num_cores"
    seq = hf.ValueSequence(path=path, values=[0, 1], nesting_order=0)
    assert seq.path == path
    assert seq.labelled_type is None
    assert seq.normalised_path == "resources.main.num_cores"
    assert seq.normalised_inputs_path is None
    assert seq.path_type == "resources"
    assert seq.input_type is None
    assert seq.input_path is None
    assert seq.resource_scope == "main"


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_value_sequence_object_values_during_workflow_init(null_config, tmp_path, store):
    p1 = hf.Parameter("p1c")
    s1 = hf.TaskSchema(objective="t1", inputs=[hf.SchemaInput(parameter=p1)])
    obj = P1(a=101)
    seq = hf.ValueSequence(path="inputs.p1c", values=[obj], nesting_order=0)
    values_exp = [P1(a=101, d=None)]

    t1 = hf.Task(
        schema=[s1],
        sequences=[seq],
    )
    # before workflow initialisation:
    assert seq.values == values_exp

    wk = hf.Workflow.from_template_data(
        tasks=[],
        path=tmp_path,
        template_name="temp",
        store=store,
    )

    with wk.batch_update():
        wk.add_task(t1)
        # after workflow initialisation but before store commit:
        assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp

    # after initialisation and store commit:
    assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_value_sequence_object_values_class_method_during_workflow_init(
    null_config, tmp_path, store
):
    p1 = hf.Parameter("p1c")
    s1 = hf.TaskSchema(objective="t1", inputs=[hf.SchemaInput(parameter=p1)])
    obj = P1.from_data(b=50, c=51)
    seq = hf.ValueSequence(path="inputs.p1c", values=[obj], nesting_order=0)
    values_exp = [P1(a=101, d=None)]

    t1 = hf.Task(
        schema=[s1],
        sequences=[seq],
    )
    # before workflow initialisation:
    assert seq.values == values_exp

    wk = hf.Workflow.from_template_data(
        tasks=[],
        path=tmp_path,
        template_name="temp",
        store=store,
    )

    with wk.batch_update():
        wk.add_task(t1)
        # after workflow initialisation but before store commit:
        assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp

    # after initialisation and store commit:
    assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_value_sequence_object_values_named_class_method_during_workflow_init(
    null_config, tmp_path, store
):
    p1 = hf.Parameter("p1c")
    s1 = hf.TaskSchema(objective="t1", inputs=[hf.SchemaInput(parameter=p1)])
    data = {"b": 50, "c": 51}
    seq = hf.ValueSequence(
        path="inputs.p1c", values=[data], nesting_order=0, value_class_method="from_data"
    )
    values_exp = [data]

    t1 = hf.Task(
        schema=[s1],
        sequences=[seq],
    )
    # before workflow initialisation:
    assert seq.values == values_exp

    wk = hf.Workflow.from_template_data(
        tasks=[],
        path=tmp_path,
        template_name="temp",
        store=store,
    )

    with wk.batch_update():
        wk.add_task(t1)
        # after workflow initialisation but before store commit:
        assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp

    # after initialisation and store commit:
    assert wk.tasks[0].template.element_sets[0].sequences[0].values == values_exp


def test_nesting_order_two_seqs_parallel(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=0),
            hf.ValueSequence("inputs.p2", values=["c", "d"], nesting_order=0),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 2
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "b", "p2": "d"}


def test_nesting_order_two_seqs_parallel_decimal_equiv(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=1.0),
            hf.ValueSequence("inputs.p2", values=["c", "d"], nesting_order=1.0),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 2
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "b", "p2": "d"}


def test_nesting_order_two_seqs_nested(
    null_config,
    tmp_path,
):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=0),
            hf.ValueSequence("inputs.p2", values=["c", "d"], nesting_order=1),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 4
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "a", "p2": "d"}
    assert wk.tasks.test.elements[2].get("inputs") == {"p1": "b", "p2": "c"}
    assert wk.tasks.test.elements[3].get("inputs") == {"p1": "b", "p2": "d"}


def test_nesting_order_two_seqs_default_nesting_order(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"]),
            hf.ValueSequence("inputs.p2", values=["c", "d"]),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 2
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "b", "p2": "d"}


def test_raise_nesting_order_two_seqs_default_nesting_order(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"]),
            hf.ValueSequence("inputs.p2", values=["c", "d", "e"]),
        ],
    )
    with pytest.raises(ValueError):
        wk = hf.Workflow.from_template_data(
            template_name="test",
            tasks=[t1],
            path=tmp_path,
        )


def test_raise_nesting_order_two_seqs_default_nesting_order_decimal(
    null_config, tmp_path
):
    ts = hf.TaskSchema(
        objective="test", inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2")]
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=0.0),
            hf.ValueSequence("inputs.p2", values=["c", "d", "e"], nesting_order=0.0),
        ],
    )
    with pytest.raises(ValueError):
        wk = hf.Workflow.from_template_data(
            template_name="test",
            tasks=[t1],
            path=tmp_path,
        )


def test_nesting_order_three_seqs_decimal(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test",
        inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2"), hf.SchemaInput("p3")],
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=0),
            hf.ValueSequence("inputs.p2", values=["c", "d", "e"], nesting_order=1),
            hf.ValueSequence(
                "inputs.p3", values=["f", "g", "h", "i", "j", "k"], nesting_order=1.5
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 6
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c", "p3": "f"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "a", "p2": "d", "p3": "g"}
    assert wk.tasks.test.elements[2].get("inputs") == {"p1": "a", "p2": "e", "p3": "h"}
    assert wk.tasks.test.elements[3].get("inputs") == {"p1": "b", "p2": "c", "p3": "i"}
    assert wk.tasks.test.elements[4].get("inputs") == {"p1": "b", "p2": "d", "p3": "j"}
    assert wk.tasks.test.elements[5].get("inputs") == {"p1": "b", "p2": "e", "p3": "k"}


def test_nesting_order_three_seqs_all_decimal(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="test",
        inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2"), hf.SchemaInput("p3")],
    )
    t1 = hf.Task(
        schema=ts,
        sequences=[
            hf.ValueSequence("inputs.p1", values=["a", "b"], nesting_order=0.5),
            hf.ValueSequence("inputs.p2", values=["c", "d", "e"], nesting_order=1.2),
            hf.ValueSequence(
                "inputs.p3", values=["f", "g", "h", "i", "j", "k"], nesting_order=1.5
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        path=tmp_path,
    )
    assert wk.tasks.test.num_elements == 6
    assert wk.tasks.test.elements[0].get("inputs") == {"p1": "a", "p2": "c", "p3": "f"}
    assert wk.tasks.test.elements[1].get("inputs") == {"p1": "a", "p2": "d", "p3": "g"}
    assert wk.tasks.test.elements[2].get("inputs") == {"p1": "a", "p2": "e", "p3": "h"}
    assert wk.tasks.test.elements[3].get("inputs") == {"p1": "b", "p2": "c", "p3": "i"}
    assert wk.tasks.test.elements[4].get("inputs") == {"p1": "b", "p2": "d", "p3": "j"}
    assert wk.tasks.test.elements[5].get("inputs") == {"p1": "b", "p2": "e", "p3": "k"}


def test_demo_data_values(null_config):
    name = "text_file.txt"
    assert hf.ValueSequence(
        path="inputs.p1", values=[f"<<demo_data_file:{name}>>"]
    ).values[0] == str(hf.demo_data_cache_dir.joinpath(name))


def test_from_linear_space(null_config):
    seq = hf.ValueSequence.from_linear_space(path="inputs.p1", start=0, stop=1, num=6)
    assert np.allclose(seq.values, [0, 0.2, 0.4, 0.6, 0.8, 1.0])


def test_from_rectangle(null_config):
    kwargs = dict(
        path="inputs.p1",
        start=[0, 0],
        stop=[1, 1],
        num=[2, 2],
    )
    seq_coord_0 = hf.ValueSequence.from_rectangle(**kwargs, coord=0)
    seq_coord_1 = hf.ValueSequence.from_rectangle(**kwargs, coord=1)

    assert np.allclose(seq_coord_0.values, [0, 1, 0, 1])
    assert np.allclose(seq_coord_1.values, [1, 1, 0, 0])


def test_from_rectangle_coord_none(null_config):
    kwargs = dict(
        path="inputs.p1",
        start=[0, 0],
        stop=[1, 1],
        num=[2, 2],
    )
    seq = hf.ValueSequence.from_rectangle(**kwargs)
    assert np.allclose(seq.values, [[0, 1], [1, 1], [0, 0], [1, 0]])



from pathlib import Path
import shutil
import signal
from typing import Callable, Dict, List, Optional, Tuple

import psutil
from hpcflow.sdk.submission.jobscript_info import JobscriptElementState

from hpcflow.sdk.submission.schedulers import NullScheduler
from hpcflow.sdk.submission.shells.base import Shell


class DirectScheduler(NullScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def process_resources(cls, resources, scheduler_config: Dict) -> None:
        """Perform scheduler-specific processing to the element resources.

        Note: this mutates `resources`.

        """
        return

    def get_submit_command(
        self,
        shell: Shell,
        js_path: str,
        deps: List[Tuple],
    ) -> List[str]:
        return shell.get_direct_submit_command(js_path)

    @staticmethod
    def _kill_processes(
        procs: List[psutil.Process],
        sig=signal.SIGTERM,
        timeout=None,
        on_terminate=None,
    ):
        all_procs = []
        for i in procs:
            all_procs.append(i)
            all_procs.extend(i.children(recursive=True))

        for i in all_procs:
            try:
                i.send_signal(sig)
            except psutil.NoSuchProcess:
                pass
        gone, alive = psutil.wait_procs(all_procs, timeout=timeout, callback=on_terminate)
        for p in alive:
            p.kill()

    @staticmethod
    def _get_jobscript_processes(js_refs: List[Tuple[int, List[str]]]):
        procs = []
        for p_id, p_cmdline in js_refs:
            try:
                proc_i = psutil.Process(p_id)
            except psutil.NoSuchProcess:
                # process might have completed already
                continue
            if proc_i.cmdline() == p_cmdline:
                # additional check this is the same process that we submitted
                procs.append(proc_i)
        return procs

    @classmethod
    def wait_for_jobscripts(
        cls,
        js_refs: List[Tuple[int, List[str]]],
        callback: Optional[Callable] = None,
    ) -> None:
        """Wait until the specified jobscripts have completed."""
        procs = cls._get_jobscript_processes(js_refs)
        (gone, alive) = psutil.wait_procs(procs, callback=callback)
        assert not alive
        return gone

    def get_job_state_info(
        self,
        js_refs: List[Tuple[int, List[str]]],
        num_js_elements: int,
    ) -> Dict[int, Dict[int, JobscriptElementState]]:
        """Query the scheduler to get the states of all of this user's jobs, optionally
        filtering by specified job IDs.

        Jobs that are not in the scheduler's status output will not appear in the output
        of this method."""
        info = {}
        for p_id, p_cmdline in js_refs:
            is_active = self.is_jobscript_active(p_id, p_cmdline)
            if is_active:
                # as far as the "scheduler" is concerned, all elements are running:
                info[p_id] = {
                    i: JobscriptElementState.running for i in range(num_js_elements)
                }

        return info

    def cancel_jobs(
        self,
        js_refs: List[Tuple[int, List[str]]],
        jobscripts: List = None,
    ):
        def callback(proc):
            try:
                js = js_proc_id[proc.pid]
            except KeyError:
                # child process of one of the jobscripts
                self.app.submission_logger.debug(
                    f"jobscript child process ({proc.pid}) killed"
                )
                return
            print(
                f"Jobscript {js.index} from submission {js.submission.index} "
                f"terminated (user-initiated cancel) with exit code {proc.returncode}."
            )

        procs = self._get_jobscript_processes(js_refs)
        self.app.submission_logger.info(
            f"cancelling {self.__class__.__name__} jobscript processes: {procs}."
        )
        js_proc_id = {i.pid: jobscripts[idx] for idx, i in enumerate(procs)}
        self._kill_processes(procs, timeout=3, on_terminate=callback)
        self.app.submission_logger.info(f"jobscripts cancel command executed.")

    def is_jobscript_active(self, process_ID: int, process_cmdline: List[str]):
        """Query if a jobscript is running.

        Note that a "running" jobscript might be waiting on upstream jobscripts to
        complete.

        """
        try:
            proc = psutil.Process(process_ID)
        except psutil.NoSuchProcess:
            return False

        if proc.cmdline() == process_cmdline:
            return True
        else:
            return False


class DirectPosix(DirectScheduler):
    _app_attr = "app"
    DEFAULT_SHELL_EXECUTABLE = "/bin/bash"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DirectWindows(DirectScheduler):
    _app_attr = "app"
    DEFAULT_SHELL_EXECUTABLE = "powershell.exe"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_submit_command(
        self, shell: Shell, js_path: str, deps: List[Tuple]
    ) -> List[str]:
        cmd = super().get_submit_command(shell, js_path, deps)
        # `Start-Process` (see `Jobscript._launch_direct_js_win`) seems to resolve the
        # executable, which means the process's `cmdline` might look different to what we
        # record; so let's resolve it ourselves:
        cmd[0] = shutil.which(cmd[0])
        return cmd



from pathlib import Path
import sys
import time
from typing import Any, List, Tuple


class NullScheduler:
    DEFAULT_SHELL_ARGS = ""
    DEFAULT_SHEBANG_ARGS = ""

    def __init__(
        self,
        shell_args=None,
        shebang_args=None,
        options=None,
    ):
        self.shebang_args = shebang_args or self.DEFAULT_SHEBANG_ARGS
        self.shell_args = shell_args or self.DEFAULT_SHELL_ARGS
        self.options = options or []

    @property
    def unique_properties(self):
        return (self.__class__.__name__,)

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        else:
            return self.__dict__ == other.__dict__

    def get_version_info(self):
        return {}

    def parse_submission_output(self, stdout: str) -> None:
        return None

    @staticmethod
    def is_num_cores_supported(num_cores, core_range: List[int]):
        step = core_range[1] if core_range[1] is not None else 1
        upper = core_range[2] + 1 if core_range[2] is not None else sys.maxsize
        return num_cores in range(core_range[0], upper, step)


class Scheduler(NullScheduler):
    DEFAULT_LOGIN_NODES_CMD = None
    DEFAULT_LOGIN_NODE_MATCH = "*login*"

    def __init__(
        self,
        submit_cmd=None,
        show_cmd=None,
        del_cmd=None,
        js_cmd=None,
        login_nodes_cmd=None,
        array_switch=None,
        array_item_var=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.submit_cmd = submit_cmd or self.DEFAULT_SUBMIT_CMD
        self.show_cmd = show_cmd or self.DEFAULT_SHOW_CMD
        self.del_cmd = del_cmd or self.DEFAULT_DEL_CMD
        self.js_cmd = js_cmd or self.DEFAULT_JS_CMD
        self.login_nodes_cmd = login_nodes_cmd or self.DEFAULT_LOGIN_NODES_CMD
        self.array_switch = array_switch or self.DEFAULT_ARRAY_SWITCH
        self.array_item_var = array_item_var or self.DEFAULT_ARRAY_ITEM_VAR

    @property
    def unique_properties(self):
        return (self.__class__.__name__, self.submit_cmd, self.show_cmd, self.del_cmd)

    def format_switch(self, switch):
        return f"{self.js_cmd} {switch}"

    def is_jobscript_active(self, job_ID: str):
        """Query if a jobscript is running/pending."""
        return bool(self.get_job_state_info([job_ID]))

    def wait_for_jobscripts(self, js_refs: List[Any]) -> None:
        while js_refs:
            info = self.get_job_state_info(js_refs)
            print(info)
            if not info:
                break
            js_refs = list(info.keys())
            time.sleep(2)



import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import P1_parameter_cls as P1

# note: when testing the frozen app, we might not have MatFlow installed in the built in
# python_env MatFlow environment, so we should skip these tests.


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_direct_in_direct_out(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_direct_sub_param_in_direct_out(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_sub_param_in_direct_out.py>>",
                script_data_in={"p1.a": "direct"},
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = {"a": 101}
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_val["a"] + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_direct_in_direct_out_single_label(null_config, tmp_path):
    """This uses the same test script as the `test_script_direct_in_direct_out` test;
    single labels are trivial and need not be referenced in the script."""
    p1_label = "one"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"), labels={p1_label: {}})],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={f"p1[{p1_label}]": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_direct_in_direct_out_labels(null_config, tmp_path):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out_labels.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_1_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": p1_1_val,
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_1_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_in_json_out(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_out.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_in_json_out_labels(null_config, tmp_path):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_out_labels.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_1_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": p1_1_val,
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_1_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_sub_param_in_json_out_labels(null_config, tmp_path):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_sub_param_in_json_out_labels.py>>",
                script_data_in={"p1[one].a": "json", "p1[two]": "json"},
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    a_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": {"a": a_val},
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == a_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_and_direct_in_json_out(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("p1")),
            hf.SchemaInput(parameter=hf.Parameter("p2")),
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p3"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_and_direct_in_json_out.py>>",
                script_data_in={"p1": "json", "p2": "direct"},
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    p2_val = 201
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val, "p2": p2_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p3.value == p1_val + p2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_in_json_and_direct_out(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[
            hf.SchemaInput(parameter=hf.Parameter("p2")),
            hf.SchemaOutput(parameter=hf.Parameter("p3")),
        ],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_and_direct_out.py>>",
                script_data_in="json",
                script_data_out={"p2": "json", "p3": "direct"},
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == p1_val + 100
    assert wk.tasks[0].elements[0].outputs.p3.value == p1_val + 200


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_in_obj(null_config, tmp_path):
    """Use a custom JSON dumper defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_obj.py>>",
                script_data_in="json",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    a_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1c": P1(a=a_val)})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == a_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_hdf5_in_obj(null_config, tmp_path):
    """Use a custom HDF5 dumper defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_in_obj.py>>",
                script_data_in="hdf5",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    a_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1c": P1(a=a_val)})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p2.value == a_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_json_out_obj(null_config, tmp_path):
    """Use a custom JSON saver defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_out_obj.py>>",
                script_data_in="direct",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    p1_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p1c.value == P1(a=p1_val + 100)


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_script_hdf5_out_obj(null_config, tmp_path):
    """Use a custom HDF5 saver defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_out_obj.py>>",
                script_data_in="direct",
                script_data_out="hdf5",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    p1_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="main_script_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks[0].elements[0].outputs.p1c.value == P1(a=p1_val + 100)



import json


def main_script_test_json_and_direct_in_json_out(p2, _input_files, _output_files):
    # read inputs
    with _input_files["json"].open("rt") as fh:
        p1 = int(json.load(fh)["p1"])

    # process
    p3 = p1 + p2

    # save outputs
    with _output_files["json"].open("wt") as fh:
        json.dump({"p3": p3}, fh)



import pytest

from valida.conditions import Value


from hpcflow.app import app as hf
from hpcflow.sdk.core.errors import LoopAlreadyExistsError
from hpcflow.sdk.core.test_utils import P1_parameter_cls, make_workflow


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_loop_tasks_obj_insert_ID_equivalence(tmp_path, store):
    wk_1 = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    lp_0 = hf.Loop(tasks=[wk_1.tasks.t1], num_iterations=2)
    lp_1 = hf.Loop(tasks=[0], num_iterations=2)
    assert lp_0.task_insert_IDs == lp_1.task_insert_IDs


def test_raise_on_add_loop_same_name(tmp_path):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"], [{"p2": None}, ("p2",), "t2"]],
        local_inputs={0: ("p1",), 1: ("p2",)},
        path=tmp_path,
        store="json",
    )
    lp_0 = hf.Loop(name="my_loop", tasks=[0], num_iterations=2)
    lp_1 = hf.Loop(name="my_loop", tasks=[1], num_iterations=2)

    wk.add_loop(lp_0)
    with pytest.raises(LoopAlreadyExistsError):
        wk.add_loop(lp_1)


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_data_idx_single_task_single_element_single_parameter_three_iters(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[wk.tasks.t1], num_iterations=3))
    iter_0, iter_1, iter_2 = wk.tasks.t1.elements[0].iterations

    p1_idx_i0_out = iter_0.get_data_idx()["outputs.p1"]
    p1_idx_i1_in = iter_1.get_data_idx()["inputs.p1"]
    p1_idx_i1_out = iter_1.get_data_idx()["outputs.p1"]
    p1_idx_i2_in = iter_2.get_data_idx()["inputs.p1"]

    assert p1_idx_i0_out == p1_idx_i1_in and p1_idx_i1_out == p1_idx_i2_in


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_EARs_initialised_single_task_single_element_single_parameter_three_iters(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[wk.tasks.t1], num_iterations=3))
    iter_0, iter_1, iter_2 = wk.tasks.t1.elements[0].iterations
    assert iter_0.EARs_initialised and iter_1.EARs_initialised and iter_2.EARs_initialised


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_data_idx_single_task_multi_element_single_parameter_three_iters(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_sequences={0: [("inputs.p1", 2, 0)]},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[wk.tasks.t1], num_iterations=3))
    e0_iter_0, e0_iter_1, e0_iter_2 = wk.tasks.t1.elements[0].iterations
    e1_iter_0, e1_iter_1, e1_iter_2 = wk.tasks.t1.elements[1].iterations

    e0_p1_idx_i0_out = e0_iter_0.get_data_idx()["outputs.p1"]
    e0_p1_idx_i1_in = e0_iter_1.get_data_idx()["inputs.p1"]
    e0_p1_idx_i1_out = e0_iter_1.get_data_idx()["outputs.p1"]
    e0_p1_idx_i2_in = e0_iter_2.get_data_idx()["inputs.p1"]

    e1_p1_idx_i0_out = e1_iter_0.get_data_idx()["outputs.p1"]
    e1_p1_idx_i1_in = e1_iter_1.get_data_idx()["inputs.p1"]
    e1_p1_idx_i1_out = e1_iter_1.get_data_idx()["outputs.p1"]
    e1_p1_idx_i2_in = e1_iter_2.get_data_idx()["inputs.p1"]

    assert (
        e0_p1_idx_i0_out == e0_p1_idx_i1_in
        and e0_p1_idx_i1_out == e0_p1_idx_i2_in
        and e1_p1_idx_i0_out == e1_p1_idx_i1_in
        and e1_p1_idx_i1_out == e1_p1_idx_i2_in
    )


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_data_idx_multi_task_single_element_single_parameter_two_iters(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[
            [{"p1": None}, ("p1",), "t1"],
            [{"p1": None}, ("p1",), "t2"],
            [{"p1": None}, ("p1",), "t3"],
        ],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[0, 1, 2], num_iterations=2))
    t1_iter_0, t1_iter_1 = wk.tasks.t1.elements[0].iterations
    t2_iter_0, t2_iter_1 = wk.tasks.t2.elements[0].iterations
    t3_iter_0, t3_iter_1 = wk.tasks.t3.elements[0].iterations

    in_key = "inputs.p1"
    out_key = "outputs.p1"

    t1_i0_p1_idx_out = t1_iter_0.get_data_idx()[out_key]
    t2_i0_p1_idx_in = t2_iter_0.get_data_idx()[in_key]
    t2_i0_p1_idx_out = t2_iter_0.get_data_idx()[out_key]
    t3_i0_p1_idx_in = t3_iter_0.get_data_idx()[in_key]
    t3_i0_p1_idx_out = t3_iter_0.get_data_idx()[out_key]

    t1_i1_p1_idx_in = t1_iter_1.get_data_idx()[in_key]
    t1_i1_p1_idx_out = t1_iter_1.get_data_idx()[out_key]
    t2_i1_p1_idx_in = t2_iter_1.get_data_idx()[in_key]
    t2_i1_p1_idx_out = t2_iter_1.get_data_idx()[out_key]
    t3_i1_p1_idx_in = t3_iter_1.get_data_idx()[in_key]

    assert (
        t1_i0_p1_idx_out == t2_i0_p1_idx_in
        and t2_i0_p1_idx_out == t3_i0_p1_idx_in
        and t3_i0_p1_idx_out == t1_i1_p1_idx_in
        and t1_i1_p1_idx_out == t2_i1_p1_idx_in
        and t2_i1_p1_idx_out == t3_i1_p1_idx_in
    )


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_data_idx_single_task_single_element_single_parameter_three_iters_non_iterable_param(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(
        hf.Loop(tasks=[wk.tasks.t1], num_iterations=3, non_iterable_parameters=["p1"])
    )
    iter_0, iter_1, iter_2 = wk.tasks.t1.elements[0].iterations

    p1_idx_i0_out = iter_0.get_data_idx()["outputs.p1"]
    p1_idx_i1_in = iter_1.get_data_idx()["inputs.p1"]
    p1_idx_i1_out = iter_1.get_data_idx()["outputs.p1"]
    p1_idx_i2_in = iter_2.get_data_idx()["inputs.p1"]

    assert p1_idx_i0_out != p1_idx_i1_in and p1_idx_i1_out != p1_idx_i2_in


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_iterable_parameters(tmp_path, store):
    wk = make_workflow(
        schemas_spec=[
            [{"p1": None, "p2": None}, ("p1", "p2"), "t1"],
            [{"p1": None}, ("p1",), "t2"],
            [{"p1": None, "p2": None}, ("p1", "p2"), "t3"],
        ],
        local_inputs={0: ("p1", "p2"), 1: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[0, 1, 2], num_iterations=2))
    assert dict(sorted(wk.loops[0].iterable_parameters.items(), key=lambda x: x[0])) == {
        "p1": {"input_task": 0, "output_tasks": [0, 1, 2]},
        "p2": {"input_task": 0, "output_tasks": [0, 2]},
    }


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_wk_loop_input_sources_including_local_single_element_two_iters(tmp_path, store):
    wk = make_workflow(
        schemas_spec=[
            [{"p1": None, "p2": None}, ("p1", "p2"), "t1"],
            [{"p1": None}, ("p1",), "t2"],
            [{"p1": None, "p2": None}, ("p1", "p2"), "t3"],
        ],
        local_inputs={0: ("p1", "p2"), 1: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(tasks=[0, 1, 2], num_iterations=2))

    t2_iter_0 = wk.tasks.t2.elements[0].iterations[0]
    t3_iter_0 = wk.tasks.t3.elements[0].iterations[0]
    t1_iter_1 = wk.tasks.t1.elements[0].iterations[1]
    t2_iter_1 = wk.tasks.t2.elements[0].iterations[1]

    t3_p1_i0_out = t3_iter_0.get_data_idx()["outputs.p1"]
    t3_p2_i0_out = t3_iter_0.get_data_idx()["outputs.p2"]

    t1_p1_i1_in = t1_iter_1.get_data_idx()["inputs.p1"]
    t1_p2_i1_in = t1_iter_1.get_data_idx()["inputs.p2"]

    # local input defined in task 2 is not an input task of the iterative parameter p1,
    # so it is sourced in all iterations from the original local input:
    t2_p1_i0_in = t2_iter_0.get_data_idx()["inputs.p1"]
    t2_p1_i1_in = t2_iter_1.get_data_idx()["inputs.p1"]

    assert (
        t3_p1_i0_out == t1_p1_i1_in
        and t3_p2_i0_out == t1_p2_i1_in
        and t2_p1_i0_in == t2_p1_i1_in
    )


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_get_iteration_task_pathway_single_task_single_element_three_iters(
    tmp_path, store
):
    wk = make_workflow(
        schemas_spec=[[{"p1": None}, ("p1",), "t1"]],
        local_inputs={0: ("p1",)},
        path=tmp_path,
        store=store,
    )
    wk.add_loop(hf.Loop(name="loop_0", tasks=[wk.tasks.t1], num_iterations=3))

    assert wk.get_iteration_task_pathway() == [
        (0, {"loop_0": 0}),
        (0, {"loop_0": 1}),
        (0, {"loop_0": 2}),
    ]


def test_wk_loop_input_sources_including_non_iteration_task_source(null_config, tmp_path):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1>> + 100))",
                        stdout="<<int(parameter:p2)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    ts2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput("p2"), hf.SchemaInput("p3")],
        outputs=[hf.SchemaOutput("p4")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p2>> + <<parameter:p3>>))",
                        stdout="<<int(parameter:p4)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    ts3 = hf.TaskSchema(
        objective="t3",
        inputs=[hf.SchemaInput("p3"), hf.SchemaInput("p4")],
        outputs=[hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p3>> + <<parameter:p4>>))",
                        stdout="<<int(parameter:p3)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[
            hf.Task(schema=ts1, inputs={"p1": 101}),
            hf.Task(schema=ts2, inputs={"p3": 301}),
            hf.Task(schema=ts3),
        ],
    )
    wk.add_loop(hf.Loop(tasks=[1, 2], num_iterations=2))
    t1 = wk.tasks.t1.elements[0].iterations[0].get_data_idx()
    t2_iter_0 = wk.tasks.t2.elements[0].iterations[0].get_data_idx()
    t3_iter_0 = wk.tasks.t3.elements[0].iterations[0].get_data_idx()
    t2_iter_1 = wk.tasks.t2.elements[0].iterations[1].get_data_idx()
    t3_iter_1 = wk.tasks.t3.elements[0].iterations[1].get_data_idx()

    assert t2_iter_0["inputs.p2"] == t2_iter_1["inputs.p2"] == t1["outputs.p2"]
    assert t3_iter_0["inputs.p3"] == t2_iter_0["inputs.p3"]
    assert t3_iter_0["inputs.p4"] == t2_iter_0["outputs.p4"]
    assert t3_iter_1["inputs.p3"] == t2_iter_1["inputs.p3"]
    assert t3_iter_1["inputs.p4"] == t2_iter_1["outputs.p4"]
    assert t2_iter_1["inputs.p3"] == t3_iter_0["outputs.p3"]


def test_wk_loop_input_sources_default(null_config, tmp_path):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2", default_value=2)],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1>> + <<parameter:p2>>))",
                        stdout="<<int(parameter:p1)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[hf.Task(schema=ts1, inputs={"p1": 101})],
    )
    wk.add_loop(hf.Loop(tasks=[0], num_iterations=2))
    t1_iter_0 = wk.tasks.t1.elements[0].iterations[0].get_data_idx()
    t1_iter_1 = wk.tasks.t1.elements[0].iterations[1].get_data_idx()

    assert t1_iter_0["inputs.p2"] == t1_iter_1["inputs.p2"]


def test_wk_loop_input_sources_iterable_param_default(null_config, tmp_path):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1", default_value=1)],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1>> + 10))",
                        stdout="<<int(parameter:p1)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[hf.Task(schema=ts1, inputs={"p1": 101})],
    )
    wk.add_loop(hf.Loop(tasks=[0], num_iterations=3))
    # first iteration should be the default value, second and third iterations should
    # be from previous iteration outputs:
    t1_iter_0 = wk.tasks.t1.elements[0].iterations[0].get_data_idx()
    t1_iter_1 = wk.tasks.t1.elements[0].iterations[1].get_data_idx()
    t1_iter_2 = wk.tasks.t1.elements[0].iterations[2].get_data_idx()

    assert t1_iter_0["inputs.p1"] != t1_iter_1["inputs.p1"]
    assert t1_iter_1["inputs.p1"] != t1_iter_2["inputs.p1"]
    assert t1_iter_1["inputs.p1"] == t1_iter_0["outputs.p1"]
    assert t1_iter_2["inputs.p1"] == t1_iter_1["outputs.p1"]


def test_wk_loop_input_sources_iterable_param_default_conditional_action(
    null_config, tmp_path
):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput("p1", default_value=1),
            hf.SchemaInput("p2", default_value=None),
        ],
        outputs=[hf.SchemaOutput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1>> + 10))",
                        stdout="<<int(parameter:p1)>>",
                    )
                ],
                environments=[act_env],
            ),
            hf.Action(
                commands=[hf.Command("Write-Output ((<<parameter:p2>> + 10))")],
                environments=[act_env],
                rules=[
                    hf.ActionRule(path="inputs.p2", condition=Value.not_equal_to(None))
                ],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[hf.Task(schema=ts1, inputs={"p1": 101})],
    )
    wk.add_loop(hf.Loop(tasks=[0], num_iterations=3))
    # first iteration should be the default value, second and third iterations should
    # be from previous iteration outputs:
    t1_iter_0 = wk.tasks.t1.elements[0].iterations[0].get_data_idx()
    t1_iter_1 = wk.tasks.t1.elements[0].iterations[1].get_data_idx()
    t1_iter_2 = wk.tasks.t1.elements[0].iterations[2].get_data_idx()

    assert t1_iter_0["inputs.p1"] != t1_iter_1["inputs.p1"]
    assert t1_iter_1["inputs.p1"] != t1_iter_2["inputs.p1"]
    assert t1_iter_1["inputs.p1"] == t1_iter_0["outputs.p1"]
    assert t1_iter_2["inputs.p1"] == t1_iter_1["outputs.p1"]


def test_wk_loop_input_sources_including_non_iteration_task_source_with_groups(
    null_config, tmp_path
):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1>> + 100))",
                        stdout="<<int(parameter:p2)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    ts2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput("p2"), hf.SchemaInput("p3")],
        outputs=[hf.SchemaOutput("p4")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p2>> + <<parameter:p3>>))",
                        stdout="<<int(parameter:p4)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    ts3 = hf.TaskSchema(
        objective="t3",
        inputs=[
            hf.SchemaInput("p3", labels={"": {"group": "my_group"}}),
            hf.SchemaInput("p4", labels={"": {"group": "my_group"}}),
        ],
        outputs=[hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<sum(parameter:p3)>> + <<sum(parameter:p4)>>))",
                        stdout="<<int(parameter:p3)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[
            hf.Task(schema=ts1, inputs={"p1": 101}),
            hf.Task(
                schema=ts2,
                sequences=[hf.ValueSequence(path="inputs.p3", values=[301, 302])],
                groups=[hf.ElementGroup(name="my_group")],
            ),
            hf.Task(schema=ts3),
        ],
    )
    wk.add_loop(hf.Loop(tasks=[1, 2], num_iterations=2))

    t2_elem_0_iter_0 = wk.tasks.t2.elements[0].iterations[0].get_data_idx()
    t2_elem_1_iter_0 = wk.tasks.t2.elements[1].iterations[0].get_data_idx()
    t2_elem_0_iter_1 = wk.tasks.t2.elements[0].iterations[1].get_data_idx()
    t2_elem_1_iter_1 = wk.tasks.t2.elements[1].iterations[1].get_data_idx()

    t3_iter_0 = wk.tasks.t3.elements[0].iterations[0].get_data_idx()
    t3_iter_1 = wk.tasks.t3.elements[0].iterations[1].get_data_idx()
    assert len(t3_iter_0["inputs.p3"]) == len(t3_iter_1["inputs.p3"]) == 2
    assert len(t3_iter_0["inputs.p4"]) == len(t3_iter_1["inputs.p4"]) == 2
    assert t3_iter_0["inputs.p3"] == [
        t2_elem_0_iter_0["inputs.p3"],
        t2_elem_1_iter_0["inputs.p3"],
    ]
    assert t3_iter_0["inputs.p4"] == [
        t2_elem_0_iter_0["outputs.p4"],
        t2_elem_1_iter_0["outputs.p4"],
    ]
    assert t3_iter_1["inputs.p3"] == [
        t2_elem_0_iter_1["inputs.p3"],
        t2_elem_1_iter_1["inputs.p3"],
    ]
    assert t3_iter_1["inputs.p4"] == [
        t2_elem_0_iter_1["outputs.p4"],
        t2_elem_1_iter_1["outputs.p4"],
    ]


def test_loop_local_sub_parameters(null_config, tmp_path):
    act_env = hf.ActionEnvironment("null_env")
    ts1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1c")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p1c.a>> + 100))",
                        stdout="<<int(parameter:p2)>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    ts2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput("p2")],
        outputs=[hf.SchemaOutput("p1c")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "Write-Output ((<<parameter:p2>> + 100))",
                        stdout="<<parameter:p1c>>",
                    )
                ],
                environments=[act_env],
            ),
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_loop",
        path=tmp_path,
        tasks=[
            hf.Task(
                schema=ts1,
                inputs=[
                    hf.InputValue(parameter="p1c", value=P1_parameter_cls(a=101)),
                    hf.InputValue(parameter="p1c", path="d", value=9),
                ],
            ),
            hf.Task(schema=ts2),
        ],
    )
    wk.add_loop(hf.Loop(tasks=[0, 1], num_iterations=2))

    t1_iter_0 = wk.tasks.t1.elements[0].iterations[0].get_data_idx()
    t2_iter_0 = wk.tasks.t2.elements[0].iterations[0].get_data_idx()
    t1_iter_1 = wk.tasks.t1.elements[0].iterations[1].get_data_idx()
    t2_iter_1 = wk.tasks.t2.elements[0].iterations[1].get_data_idx()

    assert t2_iter_0["inputs.p2"] == t1_iter_0["outputs.p2"]
    assert t1_iter_1["inputs.p1c"] == t2_iter_0["outputs.p1c"]
    assert t2_iter_1["inputs.p2"] == t1_iter_1["outputs.p2"]
    assert t1_iter_0["inputs.p1c.d"] == t1_iter_1["inputs.p1c.d"]



from __future__ import annotations

from collections import defaultdict
import contextlib
from dataclasses import dataclass, fields
from datetime import datetime

from typing import Any, Dict, List, Optional, Tuple


class PendingChanges:
    """Class to store pending changes and merge them into a persistent store.

    Parameters
    ----------
    add_tasks
        Keys are new task IDs
    add_elem_iter_EAR_IDs
        Keys are element iteration IDs, then EAR action index, and values are EAR IDs.
        This is a list of EAR IDs to add to a given element iteration action.
    add_elem_iter_IDs
        Keys are element IDs, and values are iteration IDs to add to that element.
    add_elem_IDs
        Keys are task IDs, and values are element IDs to add to that task.
    add_parameters
        Keys are parameter indices and values are tuples whose first element is data to
        add and whose second element is the source dict for the new data.
    update_param_sources
        Keys are parameter indices and values are dict parameter sources to merge with
        existing source of that parameter.
    set_EAR_starts
        Keys are EAR IDs and values are tuples of start time, and start dir snapshot.
    set_EAR_ends
        Keys are EAR IDs and values are tuples of end time, end dir snapshot, exit
        code, and success boolean.
    """

    def __init__(self, app, store, resource_map):
        self.app = app
        self.store = store
        self.resource_map = resource_map

        self.add_tasks: Dict[int, AnySTask] = None
        self.add_loops: Dict[int, Dict] = None
        self.add_submissions: Dict[int, Dict] = None
        self.add_elements: Dict[int, AnySElement] = None
        self.add_elem_iters: Dict[int, AnySElementIter] = None
        self.add_EARs: Dict[int, AnySEAR] = None
        self.add_parameters: Dict[int, AnySParameter] = None
        self.add_files: List[Dict] = None
        self.add_template_components: Dict[str, Dict[str, Dict]] = None
        self.add_element_sets: Dict[int, Dict] = None

        self.add_elem_IDs: Dict[int, List] = None
        self.add_elem_iter_IDs: Dict[int, List] = None
        self.add_elem_iter_EAR_IDs: Dict[int, Dict[int, List]] = None
        self.add_submission_parts: Dict[int, Dict[str, List[int]]] = None

        self.set_EARs_initialised: List[int] = None
        self.set_EAR_submission_indices: Dict[int, int] = None
        self.set_EAR_skips: List[int] = None
        self.set_EAR_starts: Dict[int, Tuple[datetime, Dict], str] = None
        self.set_EAR_ends: Dict[int, Tuple[datetime, Dict, int, bool]] = None

        self.set_js_metadata: Dict[int, Dict[int, Any]] = None

        self.set_parameters: Dict[int, AnySParameter] = None

        self.update_param_sources: Dict[int, Dict] = None
        self.update_loop_indices: Dict[int, Dict] = None
        self.update_loop_num_iters: Dict[int, int] = None

        self.reset(is_init=True)  # set up initial data structures

    def __bool__(self):
        """Returns True if there are any outstanding pending items."""
        return (
            bool(self.add_tasks)
            or bool(self.add_loops)
            or bool(self.add_submissions)
            or bool(self.add_elements)
            or bool(self.add_elem_iters)
            or bool(self.add_EARs)
            or bool(self.add_elem_IDs)
            or bool(self.add_elem_iter_IDs)
            or bool(self.add_elem_iter_EAR_IDs)
            or bool(self.add_submission_parts)
            or bool(self.add_parameters)
            or bool(self.add_files)
            or bool(self.add_template_components)
            or bool(self.add_element_sets)
            or bool(self.set_EARs_initialised)
            or bool(self.set_EAR_submission_indices)
            or bool(self.set_EAR_starts)
            or bool(self.set_EAR_ends)
            or bool(self.set_EAR_skips)
            or bool(self.set_js_metadata)
            or bool(self.set_parameters)
            or bool(self.update_param_sources)
            or bool(self.update_loop_indices)
            or bool(self.update_loop_num_iters)
        )

    def where_pending(self) -> List[str]:
        return [
            k
            for k, v in self.__dict__.items()
            if k not in ("app", "store", "resource_map") and bool(v)
        ]

    @property
    def logger(self):
        return self.app.persistence_logger

    def commit_all(self):
        """Commit all pending changes to disk."""
        self.logger.info(f"committing all pending changes: {self.where_pending()}")

        if not self:
            self.logger.debug("commit: no pending changes to commit.")
            return

        for resources, methods in self.resource_map.groups.items():
            # for each resource, enter `using_resource` context manager in "update" mode:
            with contextlib.ExitStack() as stack:
                for res in resources:
                    # TODO: only enter required resources!
                    stack.enter_context(self.store.using_resource(res, "update"))
                for meth in methods:
                    getattr(self, meth)()

        assert not (self)

    def commit_tasks(self) -> None:
        """Commit pending tasks to disk."""
        if self.add_tasks:
            tasks = self.store.get_tasks_by_IDs(self.add_tasks)
            task_ids = list(self.add_tasks.keys())
            self.logger.debug(f"commit: adding pending tasks with IDs: {task_ids!r}")
            self.store._append_tasks(tasks)
            # pending element IDs that belong to pending tasks are now committed:
            self.add_elem_IDs = {
                k: v for k, v in self.add_elem_IDs.items() if k not in task_ids
            }
        self.clear_add_tasks()

    def commit_loops(self) -> None:
        """Commit pending loops to disk."""
        if self.add_loops:
            # retrieve pending loops, including pending changes to num_added_iterations:
            loops = self.store.get_loops_by_IDs(self.add_loops)
            loop_ids = list(self.add_loops.keys())
            self.logger.debug(f"commit: adding pending loops with indices {loop_ids!r}")
            self.store._append_loops(loops)
        self.clear_add_loops()

    def commit_submissions(self) -> None:
        """Commit pending submissions to disk."""
        if self.add_submissions:
            # retrieve pending submissions:
            subs = self.store.get_submissions_by_ID(self.add_submissions)
            sub_ids = list(self.add_submissions.keys())
            self.logger.debug(
                f"commit: adding pending submissions with indices {sub_ids!r}"
            )
            self.store._append_submissions(subs)
        self.clear_add_submissions()

    def commit_submission_parts(self) -> None:
        if self.add_submission_parts:
            self.logger.debug(f"commit: adding pending submission parts")
            self.store._append_submission_parts(self.add_submission_parts)
        self.clear_add_submission_parts()

    def commit_elem_IDs(self) -> None:
        # TODO: could be batched up?
        for task_ID, elem_IDs in self.add_elem_IDs.items():
            self.logger.debug(
                f"commit: adding pending element IDs to task {task_ID!r}: {elem_IDs!r}."
            )
            self.store._append_task_element_IDs(task_ID, elem_IDs)
        self.clear_add_elem_IDs()

    def commit_elements(self) -> None:
        if self.add_elements:
            elems = self.store.get_elements(self.add_elements)
            elem_ids = list(self.add_elements.keys())
            self.logger.debug(f"commit: adding pending elements with IDs: {elem_ids!r}")
            self.store._append_elements(elems)
            # pending iter IDs that belong to pending elements are now committed:
            self.add_elem_iter_IDs = {
                k: v for k, v in self.add_elem_iter_IDs.items() if k not in elem_ids
            }
        self.clear_add_elements()

    def commit_element_sets(self) -> None:
        # TODO: could be batched up?
        for task_id, es_js in self.add_element_sets.items():
            self.logger.debug(f"commit: adding pending element sets.")
            self.store._append_element_sets(task_id, es_js)
        self.clear_add_element_sets()

    def commit_elem_iter_IDs(self) -> None:
        # TODO: could be batched up?
        for elem_ID, iter_IDs in self.add_elem_iter_IDs.items():
            self.logger.debug(
                f"commit: adding pending element iteration IDs to element {elem_ID!r}: "
                f"{iter_IDs!r}."
            )
            self.store._append_elem_iter_IDs(elem_ID, iter_IDs)
        self.clear_add_elem_iter_IDs()

    def commit_elem_iters(self) -> None:
        if self.add_elem_iters:
            iters = self.store.get_element_iterations(self.add_elem_iters.keys())
            iter_ids = list(self.add_elem_iters.keys())
            self.logger.debug(
                f"commit: adding pending element iterations with IDs: {iter_ids!r}"
            )
            self.store._append_elem_iters(iters)
            # pending EAR IDs that belong to pending iters are now committed:
            self.add_elem_iter_EAR_IDs = {
                k: v for k, v in self.add_elem_iter_EAR_IDs.items() if k not in iter_ids
            }
            # pending EARs_initialised that belong to pending iters are now committed:
            self.set_EARs_initialised = [
                i for i in self.set_EARs_initialised if i not in iter_ids
            ]
        self.clear_add_elem_iters()

    def commit_elem_iter_EAR_IDs(self) -> None:
        # TODO: could be batched up?
        for iter_ID, act_EAR_IDs in self.add_elem_iter_EAR_IDs.items():
            self.logger.debug(
                f"commit: adding pending EAR IDs to element iteration {iter_ID!r}: "
                f"{dict(act_EAR_IDs)!r}."
            )
            for act_idx, EAR_IDs in act_EAR_IDs.items():
                self.store._append_elem_iter_EAR_IDs(iter_ID, act_idx, EAR_IDs)
        self.clear_add_elem_iter_EAR_IDs()

    def commit_EARs(self) -> None:
        if self.add_EARs:
            EARs = self.store.get_EARs(self.add_EARs)
            EAR_ids = list(self.add_EARs.keys())
            self.logger.debug(f"commit: adding pending EARs with IDs: {EAR_ids!r}")
            self.store._append_EARs(EARs)
            # pending start/end times/snapshots, submission indices, and skips that belong
            # to pending EARs are now committed (accounted for in `get_EARs` above):
            self.set_EAR_submission_indices = {
                k: v
                for k, v in self.set_EAR_submission_indices.items()
                if k not in EAR_ids
            }
            self.set_EAR_skips = [i for i in self.set_EAR_skips if i not in EAR_ids]
            self.set_EAR_starts = {
                k: v for k, v in self.set_EAR_starts.items() if k not in EAR_ids
            }
            self.set_EAR_ends = {
                k: v for k, v in self.set_EAR_ends.items() if k not in EAR_ids
            }

        self.clear_add_EARs()

    def commit_EARs_initialised(self) -> None:
        if self.set_EARs_initialised:
            iter_ids = self.set_EARs_initialised
            self.logger.debug(
                f"commit: setting pending `EARs_initialised` for iteration IDs: "
                f"{iter_ids!r}."
            )
            # TODO: could be batched up?
            for i in iter_ids:
                self._store._update_elem_iter_EARs_initialised(i)
        self.clear_set_EARs_initialised()

    def commit_EAR_submission_indices(self) -> None:
        # TODO: could be batched up?
        for EAR_id, sub_idx in self.set_EAR_submission_indices.items():
            self.logger.debug(
                f"commit: adding pending submission index ({sub_idx!r}) to EAR ID "
                f"{EAR_id!r}."
            )
            self.store._update_EAR_submission_index(EAR_id, sub_idx)
        self.clear_set_EAR_submission_indices()

    def commit_EAR_starts(self) -> None:
        # TODO: could be batched up?
        for EAR_id, (time, snap, hostname) in self.set_EAR_starts.items():
            self.logger.debug(
                f"commit: adding pending start time ({time!r}), run hostname "
                f"({hostname!r}), and directory snapshot to EAR ID {EAR_id!r}."
            )
            self.store._update_EAR_start(EAR_id, time, snap, hostname)
        self.clear_set_EAR_starts()

    def commit_EAR_ends(self) -> None:
        # TODO: could be batched up?
        for EAR_id, (time, snap, ext, suc) in self.set_EAR_ends.items():
            self.logger.debug(
                f"commit: adding pending end time ({time!r}), directory snapshot, "
                f"exit code ({ext!r}), and success status {suc!r} to EAR ID {EAR_id!r}."
            )
            self.store._update_EAR_end(EAR_id, time, snap, ext, suc)
        self.clear_set_EAR_ends()

    def commit_EAR_skips(self) -> None:
        # TODO: could be batched up?
        for EAR_id in self.set_EAR_skips:
            self.logger.debug(f"commit: setting EAR ID {EAR_id!r} as skipped.")
            self.store._update_EAR_skip(EAR_id)
        self.clear_set_EAR_skips()

    def commit_js_metadata(self) -> None:
        if self.set_js_metadata:
            self.logger.debug(
                f"commit: setting jobscript metadata: {self.set_js_metadata!r}"
            )
            self.store._update_js_metadata(self.set_js_metadata)
        self.clear_set_js_metadata()

    def commit_parameters(self) -> None:
        """Make pending parameters persistent."""
        if self.add_parameters:
            params = self.store.get_parameters(self.add_parameters)
            param_ids = list(self.add_parameters.keys())
            self.logger.debug(f"commit: adding pending parameters IDs: {param_ids!r}")
            self.store._append_parameters(params)
        self.clear_add_parameters()

        for param_id, (value, is_file) in self.set_parameters.items():
            # TODO: could be batched up?
            self.logger.debug(f"commit: setting value of parameter ID {param_id!r}.")
            self.store._set_parameter_value(param_id, value, is_file)
        self.clear_set_parameters()

    def commit_files(self) -> None:
        """Add pending files to the files directory."""
        if self.add_files:
            self.logger.debug(f"commit: adding pending files to the files directory.")
            self.store._append_files(self.add_files)
        self.clear_add_files()

    def commit_template_components(self) -> None:
        if self.add_template_components:
            self.logger.debug(f"commit: adding template components.")
            self.store._update_template_components(self.store.get_template_components())
        self.clear_add_template_components()

    def commit_param_sources(self) -> None:
        """Make pending changes to parameter sources persistent."""
        for param_id, src in self.update_param_sources.items():
            # TODO: could be batched up?
            self.logger.debug(f"commit: updating source of parameter ID {param_id!r}.")
            self.store._update_parameter_source(param_id, src)
        self.clear_update_param_sources()

    def commit_loop_indices(self) -> None:
        """Make pending update to element iteration loop indices persistent."""
        for iter_ID, loop_idx in self.update_loop_indices.items():
            self.logger.debug(
                f"commit: updating loop indices of iteration ID {iter_ID!r} with "
                f"{loop_idx!r}."
            )
            self.store._update_loop_index(iter_ID, loop_idx)
        self.clear_update_loop_indices()

    def commit_loop_num_iters(self) -> None:
        """Make pending update to the number of loop iterations."""
        for index, num_iters in self.update_loop_num_iters.items():
            self.logger.debug(
                f"commit: updating loop {index!r} number of iterations to {num_iters!r}."
            )
            self.store._update_loop_num_iters(index, num_iters)
        self.clear_update_loop_num_iters()

    def clear_add_tasks(self):
        self.add_tasks = {}

    def clear_add_loops(self):
        self.add_loops = {}

    def clear_add_submissions(self):
        self.add_submissions = {}

    def clear_add_submission_parts(self):
        self.add_submission_parts = defaultdict(dict)

    def clear_add_elements(self):
        self.add_elements = {}

    def clear_add_element_sets(self):
        self.add_element_sets = defaultdict(list)

    def clear_add_elem_iters(self):
        self.add_elem_iters = {}

    def clear_add_EARs(self):
        self.add_EARs = {}

    def clear_add_elem_IDs(self):
        self.add_elem_IDs = defaultdict(list)

    def clear_add_elem_iter_IDs(self):
        self.add_elem_iter_IDs = defaultdict(list)

    def clear_add_elem_iter_EAR_IDs(self):
        self.add_elem_iter_EAR_IDs = defaultdict(lambda: defaultdict(list))

    def clear_set_EARs_initialised(self):
        self.set_EARs_initialised = []

    def clear_set_EAR_submission_indices(self):
        self.set_EAR_submission_indices = {}

    def clear_set_EAR_starts(self):
        self.set_EAR_starts = {}

    def clear_set_EAR_ends(self):
        self.set_EAR_ends = {}

    def clear_set_EAR_skips(self):
        self.set_EAR_skips = []

    def clear_set_js_metadata(self):
        self.set_js_metadata = defaultdict(lambda: defaultdict(dict))

    def clear_add_parameters(self):
        self.add_parameters = {}

    def clear_add_files(self):
        self.add_files = []

    def clear_add_template_components(self):
        self.add_template_components = defaultdict(dict)

    def clear_set_parameters(self):
        self.set_parameters = {}

    def clear_update_param_sources(self):
        self.update_param_sources = {}

    def clear_update_loop_indices(self):
        self.update_loop_indices = {}

    def clear_update_loop_num_iters(self):
        self.update_loop_num_iters = {}

    def reset(self, is_init=False) -> None:
        """Clear all pending data and prepare to accept new pending data."""

        if not is_init and not self:
            # no pending changes
            return

        if not is_init:
            self.logger.info("resetting pending changes.")

        self.clear_add_tasks()
        self.clear_add_loops()
        self.clear_add_submissions()
        self.clear_add_submission_parts()
        self.clear_add_elements()
        self.clear_add_element_sets()
        self.clear_add_elem_iters()
        self.clear_add_EARs()

        self.clear_set_EARs_initialised()
        self.clear_add_elem_IDs()
        self.clear_add_elem_iter_IDs()
        self.clear_add_elem_iter_EAR_IDs()

        self.clear_add_parameters()
        self.clear_add_files()
        self.clear_add_template_components()

        self.clear_set_EAR_submission_indices()
        self.clear_set_EAR_starts()
        self.clear_set_EAR_ends()
        self.clear_set_EAR_skips()

        self.clear_set_js_metadata()
        self.clear_set_parameters()

        self.clear_update_param_sources()
        self.clear_update_loop_indices()
        self.clear_update_loop_num_iters()


@dataclass
class CommitResourceMap:
    """Map of `PendingChanges` commit method names to store resource labels, representing
    the store resources required by each commit method, for a given `PersistentStore`

    When `PendingChanges.commit_all` is called, the resources specified will be opened in
    "update" mode, for each `commit_` method.

    """

    commit_tasks: Optional[Tuple[str]] = tuple()
    commit_loops: Optional[Tuple[str]] = tuple()
    commit_submissions: Optional[Tuple[str]] = tuple()
    commit_submission_parts: Optional[Tuple[str]] = tuple()
    commit_elem_IDs: Optional[Tuple[str]] = tuple()
    commit_elements: Optional[Tuple[str]] = tuple()
    commit_element_sets: Optional[Tuple[str]] = tuple()
    commit_elem_iter_IDs: Optional[Tuple[str]] = tuple()
    commit_elem_iters: Optional[Tuple[str]] = tuple()
    commit_elem_iter_EAR_IDs: Optional[Tuple[str]] = tuple()
    commit_EARs_initialised: Optional[Tuple[str]] = tuple()
    commit_EARs: Optional[Tuple[str]] = tuple()
    commit_EAR_submission_indices: Optional[Tuple[str]] = tuple()
    commit_EAR_skips: Optional[Tuple[str]] = tuple()
    commit_EAR_starts: Optional[Tuple[str]] = tuple()
    commit_EAR_ends: Optional[Tuple[str]] = tuple()
    commit_js_metadata: Optional[Tuple[str]] = tuple()
    commit_parameters: Optional[Tuple[str]] = tuple()
    commit_files: Optional[Tuple[str]] = tuple()
    commit_template_components: Optional[Tuple[str]] = tuple()
    commit_param_sources: Optional[Tuple[str]] = tuple()
    commit_loop_indices: Optional[Tuple[str]] = tuple()
    commit_loop_num_iters: Optional[Tuple[str]] = tuple()

    def __post_init__(self):
        self.groups = self.group_by_resource()

    def group_by_resource(self) -> Dict[Tuple[str], List[str]]:
        """Return a dict whose keys are tuples of resource labels and whose values are
        lists of `PendingChanges` commit method names that require those resource.

        This grouping allows us to batch up commit methods by resource requirements, which
        in turn means we can potentially minimise e.g. the number of network requests.

        """
        groups = {}
        cur_res_group = None
        for fld in fields(self):
            res_labels = getattr(self, fld.name)

            if not cur_res_group:
                # start a new resource group: a mapping between resource labels and the
                # commit methods that require those resources:
                cur_res_group = [list(res_labels), [fld.name]]

            elif not res_labels or set(res_labels).intersection(cur_res_group[0]):
                # there is some overlap between resource labels required in the current
                # group and this commit method, so we merge resource labels and add the
                # new commit method:
                cur_res_group[0] = list(set(cur_res_group[0] + list(res_labels)))
                cur_res_group[1].append(fld.name)

            else:
                # no overlap between resource labels required in the current group and
                # those required by this commit method, so append the current group, and
                # start a new group for this commit method:
                if tuple(cur_res_group[0]) not in groups:
                    groups[tuple(cur_res_group[0])] = []
                groups[tuple(cur_res_group[0])].extend(cur_res_group[1])
                cur_res_group = [list(res_labels), [fld.name]]

        if cur_res_group:
            if tuple(cur_res_group[0]) not in groups:
                groups[tuple(cur_res_group[0])] = []

            groups[tuple(cur_res_group[0])].extend(cur_res_group[1])

        return groups



import enum


class JobscriptElementState(enum.Enum):
    """Enumeration to convey a particular jobscript element state as reported by the
    scheduler."""

    def __new__(cls, value, symbol, colour, doc=None):
        member = object.__new__(cls)
        member._value_ = value
        member.symbol = symbol
        member.colour = colour
        member.__doc__ = doc
        return member

    pending = (
        0,
        "○",
        "yellow",
        "Waiting for resource allocation.",
    )
    waiting = (
        1,
        "◊",
        "grey46",
        "Waiting for one or more dependencies to finish.",
    )
    running = (
        2,
        "●",
        "dodger_blue1",
        "Executing now.",
    )
    finished = (
        3,
        "■",
        "grey46",
        "Previously submitted but is no longer active.",
    )
    cancelled = (
        4,
        "C",
        "red3",
        "Cancelled by the user.",
    )
    errored = (
        5,
        "E",
        "red3",
        "The scheduler reports an error state.",
    )

    @property
    def rich_repr(self):
        return f"[{self.colour}]{self.symbol}[/{self.colour}]"



import json


def main_script_test_json_in_obj(_input_files):
    # read inputs
    with _input_files["json"].open("rt") as fh:
        p1c = json.load(fh)["p1c"]

    # process
    p2 = p1c["a"] + 100

    return {"p2": p2}



import re
import numpy as np
import pytest
from hpcflow.app import app as hf
from hpcflow.sdk.submission.shells import ALL_SHELLS
from hpcflow.sdk.core.test_utils import (
    P1_parameter_cls as P1,
    P1_sub_parameter_cls as P1_sub,
)


def test_get_command_line(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = 1
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value} + 100)"


@pytest.mark.parametrize("shell_args", [("powershell", "nt"), ("bash", "posix")])
def test_get_command_line_with_stdout(null_config, tmp_path, shell_args):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<parameter:p2>>",
                    ),
                ],
            )
        ],
    )
    p1_value = 1
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS[shell_args[0]][shell_args[1]]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    if shell_args == ("powershell", "nt"):
        assert cmd_str == f"$parameter_p2 = Write-Output ({p1_value} + 100)"

    elif shell_args == ("bash", "posix"):
        assert cmd_str == f"parameter_p2=`Write-Output ({p1_value} + 100)`"


def test_get_command_line_single_labelled_input(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"), labels={"one": {}})],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1[one]>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = 1
    tasks = [
        hf.Task(schema=s1, inputs=[hf.InputValue("p1", label="one", value=p1_value)])
    ]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value} + 100)"


def test_get_command_line_multiple_labelled_input(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"), multiple=True, labels={"one": {}, "two": {}}
            )
        ],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1[one]>> + <<parameter:p1[two]>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_one_value = 1
    p1_two_value = 2
    tasks = [
        hf.Task(
            schema=s1,
            inputs=[
                hf.InputValue("p1", label="one", value=p1_one_value),
                hf.InputValue("p1", label="two", value=p1_two_value),
            ],
        ),
    ]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_one_value} + {p1_two_value} + 100)"


def test_get_command_line_sub_parameter(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1.a>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = {"a": 1}
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value['a']} + 100)"


def test_get_command_line_sum(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command="Write-Output (<<sum(parameter:p1)>> + 100)"),
                ],
            )
        ],
    )
    p1_value = [1, 2, 3]
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({sum(p1_value)} + 100)"


def test_get_command_line_join(null_config, tmp_path):
    delim = ","
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command=f'Write-Output (<<join[delim="{delim}"](parameter:p1)>> + 100)'
                    ),
                ],
            )
        ],
    )
    p1_value = [1, 2, 3]
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({delim.join(str(i) for i in p1_value)} + 100)"


def test_get_command_line_sum_sub_data(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command="Write-Output (<<sum(parameter:p1.a)>> + 100)"),
                ],
            )
        ],
    )
    p1_value = {"a": [1, 2, 3]}
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({sum(p1_value['a'])} + 100)"


def test_get_command_line_join_sub_data(null_config, tmp_path):
    delim = ","
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command=f'Write-Output (<<join[delim="{delim}"](parameter:p1.a)>> + 100)'
                    ),
                ],
            )
        ],
    )
    p1_value = {"a": [1, 2, 3]}
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({delim.join(str(i) for i in p1_value['a'])} + 100)"


def test_get_command_line_parameter_value(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1c>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = P1(a=1)  # has a `CLI_format` method defined which returns `str(a)`
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.a} + 100)"


def test_get_command_line_parameter_value_join(null_config, tmp_path):
    delim = ","
    cmd = (
        f"Write-Output "
        f'<<join[delim="{delim}"](parameter:p1c.custom_CLI_format_prep(reps=4))>>'
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(command=cmd),
                ],
            )
        ],
    )
    p1_value = P1(a=4)
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())
    assert cmd_str == f"Write-Output 4,4,4,4"


def test_get_command_line_parameter_value_custom_method(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1c.custom_CLI_format()>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = P1(a=1)
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.a + 4} + 100)"


def test_get_command_line_parameter_value_custom_method_with_args(null_config, tmp_path):
    add_val = 35
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command=f"Write-Output (<<parameter:p1c.custom_CLI_format(add={add_val})>> + 100)",
                    ),
                ],
            )
        ],
    )
    p1_value = P1(a=1)
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.a + add_val} + 100)"


def test_get_command_line_parameter_value_custom_method_with_two_args(
    null_config, tmp_path
):
    add_val = 35
    sub_val = 10
    cmd = (
        f"Write-Output ("
        f"<<parameter:p1c.custom_CLI_format(add={add_val}, sub={sub_val})>> + 100)"
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[hf.Action(commands=[hf.Command(command=cmd)])],
    )
    p1_value = P1(a=1)
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.a + add_val - sub_val} + 100)"


def test_get_command_line_parameter_value_sub_object(null_config, tmp_path):
    cmd = f"Write-Output (<<parameter:p1c.sub_param>> + 100)"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[hf.Action(commands=[hf.Command(command=cmd)])],
    )
    p1_value = P1(a=1, sub_param=P1_sub(e=5))
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.sub_param.e} + 100)"


def test_get_command_line_parameter_value_sub_object_attr(null_config, tmp_path):
    cmd = f"Write-Output (" f"<<parameter:p1c.sub_param.e>> + 100)"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        actions=[hf.Action(commands=[hf.Command(command=cmd)])],
    )
    p1_value = P1(a=1, sub_param=P1_sub(e=5))
    tasks = [hf.Task(schema=s1, inputs=[hf.InputValue("p1c", value=p1_value)])]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        path=tmp_path,
        template_name="wk0",
        overwrite=True,
    )
    run = wk.tasks.t1.elements[0].iterations[0].action_runs[0]
    cmd = run.action.commands[0]
    shell = ALL_SHELLS["powershell"]["nt"]()
    cmd_str, _ = cmd.get_command_line(EAR=run, shell=shell, env=run.get_environment())

    assert cmd_str == f"Write-Output ({p1_value.sub_param.e} + 100)"


def test_process_std_stream_int(null_config):
    cmd = hf.Command(command="", stdout="<<int(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="101", stderr=False) == 101


def test_process_std_stream_stderr_int(null_config):
    cmd = hf.Command(command="", stderr="<<int(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="101", stderr=True) == 101


def test_process_std_stream_float(null_config):
    cmd = hf.Command(command="", stdout="<<float(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="3.1415", stderr=False) == 3.1415


def test_process_std_stream_bool_true(null_config):
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("true", "True", "1"):
        assert cmd.process_std_stream(name="p2", value=value, stderr=False) == True


def test_process_std_stream_bool_false(null_config):
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("false", "False", "0"):
        assert cmd.process_std_stream(name="p2", value=value, stderr=False) == False


def test_process_std_stream_bool_raise(null_config):
    cmd = hf.Command(command="", stdout="<<bool(parameter:p2)>>")
    for value in ("hi", "120", "-1"):
        with pytest.raises(ValueError):
            cmd.process_std_stream(name="p2", value=value, stderr=False)


def test_process_std_stream_list(null_config):
    cmd = hf.Command(command="", stdout="<<list(parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="1 2 3", stderr=False) == [
        "1",
        "2",
        "3",
    ]


def test_process_std_stream_list_int(null_config):
    cmd = hf.Command(command="", stdout="<<list[item_type=int](parameter:p2)>>")
    assert cmd.process_std_stream(name="p2", value="1 2 3", stderr=False) == [1, 2, 3]


def test_process_std_stream_list_delim(null_config):
    cmd = hf.Command(command="", stdout='<<list[delim=","](parameter:p2)>>')
    assert cmd.process_std_stream(name="p2", value="1,2,3", stderr=False) == [
        "1",
        "2",
        "3",
    ]


def test_process_std_stream_list_int_delim(null_config):
    cmd = hf.Command(
        command="", stdout='<<list[item_type=int, delim=","](parameter:p2)>>'
    )
    assert cmd.process_std_stream(name="p2", value="1,2,3", stderr=False) == [1, 2, 3]


def test_process_std_stream_list_float_delim_colon(null_config):
    cmd = hf.Command(
        command="", stdout='<<list[item_type=float, delim=":"](parameter:p2)>>'
    )
    assert cmd.process_std_stream(name="p2", value="1.1:2.2:3.3", stderr=False) == [
        1.1,
        2.2,
        3.3,
    ]


def test_process_std_stream_array(null_config):
    cmd = hf.Command(command="", stdout="<<array(parameter:p2)>>")
    assert np.allclose(
        cmd.process_std_stream(name="p2", value="1 2 3", stderr=False),
        np.array([1, 2, 3]),
    )


def test_process_std_stream_array_delim(null_config):
    cmd = hf.Command(command="", stdout='<<array[delim=","](parameter:p2)>>')
    assert np.allclose(
        cmd.process_std_stream(name="p2", value="1,2,3", stderr=False),
        np.array([1, 2, 3]),
    )


def test_process_std_stream_array_dtype_int(null_config):
    cmd = hf.Command(command="", stdout="<<array[item_type=int](parameter:p2)>>")
    arr = cmd.process_std_stream(name="p2", value="1 2 3", stderr=False)
    assert arr.dtype == np.dtype("int")


def test_process_std_stream_array_dtype_float(null_config):
    cmd = hf.Command(command="", stdout="<<array[item_type=float](parameter:p2)>>")
    arr = cmd.process_std_stream(name="p2", value="1 2 3", stderr=False)
    assert arr.dtype == np.dtype("float")


def test_process_std_stream_object(null_config):
    cmd = hf.Command(command="", stdout="<<parameter:p1c>>")
    a_val = 12
    assert cmd.process_std_stream(name="p1c", value=str(a_val), stderr=False) == P1(
        a=a_val
    )


def test_process_std_stream_object_kwargs(null_config):
    cmd = hf.Command(command="", stdout="<<parameter:p1c.CLI_parse(double=true)>>")
    a_val = 12
    expected = 2 * a_val
    assert cmd.process_std_stream(name="p1c", value=str(a_val), stderr=False) == P1(
        a=expected
    )


def test_get_output_types(null_config):
    cmd = hf.Command(command="", stdout="<<parameter:p1_test_123>>")
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_int(null_config):
    cmd = hf.Command(command="", stdout="<<int(parameter:p1_test_123)>>")
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_object_with_args(null_config):
    cmd = hf.Command(
        command="", stdout="<<parameter:p1_test_123.CLI_parse(double=true)>>"
    )
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_list(null_config):
    cmd = hf.Command(
        command="", stdout="<<list[item_type=int, delim=" "](parameter:p1_test_123)>>"
    )
    assert cmd.get_output_types() == {"stdout": "p1_test_123", "stderr": None}


def test_get_output_types_no_match(null_config):
    cmd = hf.Command(command="", stdout="parameter:p1_test_123")
    assert cmd.get_output_types() == {"stdout": None, "stderr": None}


def test_get_output_types_raise_with_extra_substring_start(null_config):
    cmd = hf.Command(command="", stdout="hello: <<parameter:p1_test_123>>")
    with pytest.raises(ValueError):
        cmd.get_output_types()


def test_get_output_types_raise_with_extra_substring_end(null_config):
    cmd = hf.Command(command="", stdout="<<parameter:p1_test_123>> hello")
    with pytest.raises(ValueError):
        cmd.get_output_types()


def test_extract_executable_labels(null_config):
    tests = {
        "<<executable:m1>> and <<executable:12>>": ["m1", "12"],
        "<<executable:m1>> hi": ["m1"],
        "<<executable:m1": [],
    }
    for k, v in tests.items():
        assert hf.Command._extract_executable_labels(k) == v



import pytest

from click.testing import CliRunner

from hpcflow import __version__
from hpcflow.app import app as hf


def test_version():
    runner = CliRunner()
    result = runner.invoke(hf.cli, args="--version")
    assert result.output.strip() == f"hpcFlow, version {__version__}"
